library NetCommandsLog;

{$mode objfpc}{$H+}

uses
	cmem,
	SysUtils, IniFiles,
	Windows,
	MinHook,
	ModuleUtils, mylogger;

type
	PDataStore = ^TDataStore;

	TDataStore_VMT = packed record
		method00: Pointer;
		method04: Pointer;
		method08: Pointer;
		method0C: Pointer;
		method10: Pointer;
		method14: Pointer;
		method18: Pointer;
		method1C: Pointer;
		method20: Pointer;
		GetBuffer: procedure (Dummy1, Dummy2: Pointer; this: PDataStore; out Capacity, Length: Dword; out Buffer: Pointer); register; // thiscall
		method28: Pointer;
	end;
	PDataStore_VMT = ^TDataStore_VMT;

	TDataStore = packed record
		vtable: PDataStore_VMT;
		// ...
	end;

	TNetData_FlushCommands_Method = procedure (Dummy1: Pointer; Dummy2: Pointer; this: Pointer); register; // thiscall
	TNetSendCommandStoreFunction = procedure (Dummy: Pointer; Session: Dword; DataStore: PDataStore); register; // fastcall

const
	CONFIG_PATH = './NetCommandsLog.cfg';

var
	GameBase: Pointer;
	_CNetData_FlushCommands: TNetData_FlushCommands_Method;
	orig_CNetData_FlushCommands: TNetData_FlushCommands_Method;
	_NetSendCommandStore: TNetSendCommandStoreFunction;
	orig_NetSendCommandStore: TNetSendCommandStoreFunction;

	Logger: TLogger;

function CDataStore_GetLength(this: PDataStore): Dword;
var
	Buffer: Pointer;
	Length, Capacity: Dword;
begin
	this^.vtable^.GetBuffer(nil, nil, this, Capacity, Length, Buffer);
	Result:= Length;
end;

function CNetData_GetCommandBufferReservedBytesCount(this: Pointer): Dword;
begin
	Result:= PDword(this + $2244)^;
end;

function CNetData_GetCommandBuffer(this: Pointer): PDataStore;
begin
	Result:= PDataStore(this + $1C78);
end;

function CNetData_GetCommandsSize(this: Pointer): Dword;
var
	CommandBuffer: PDataStore;
begin
	CommandBuffer:= CNetData_GetCommandBuffer(this);
	Result:= CDataStore_GetLength(CommandBuffer)
		- CNetData_GetCommandBufferReservedBytesCount(this);
end;

procedure hook_CNetData_FlushCommands({%H-}Dummy1: Pointer; {%H-}Dummy2: Pointer; this: Pointer); register;
var
	CommandsSize: Dword;
begin
	CommandsSize:= CNetData_GetCommandsSize(this);
	if CommandsSize > 0 then begin
		Logger.PrintLine('Command buffer flushed (%d bytes).', [CommandsSize]);
	end;

	orig_CNetData_FlushCommands(nil, nil, this);
end;

function BytesToHex(Bytes: PByte; Count: Integer; Delimiter: String): String;
const
	HEX_CHARS: array [0..15] of Char = (
		'0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
		'A', 'B', 'C', 'D', 'E', 'F'
	);
	BYTE_LENGTH_IN_HEX = 2;
var
	i: Integer;
	ItemSize: Integer;
begin
	Result:= '';
	ItemSize:= (BYTE_LENGTH_IN_HEX + Length(Delimiter));
	SetLength(Result, ItemSize * Count - Length(Delimiter));

	for i:= 0 to Count - 1 do begin
		Result[ItemSize * i + 1]:= HEX_CHARS[Bytes[i] div 16];
		Result[ItemSize * i + 2]:= HEX_CHARS[Bytes[i] mod 16];

		if i < Count - 1 then begin
			if Delimiter <> ''
				then Move(
					Delimiter[1],
					Result[ItemSize * i + 3],
					Length(Delimiter)
				);
		end;
	end;
end;

procedure hook_NetSendCommandStore({%H-}Dummy: Pointer; Session: Dword; DataStore: PDataStore); register;
var
	Buffer: PByte;
	Length, Capacity: Dword;
	HexBytes: String;
begin
	DataStore^.vtable^.GetBuffer(nil, nil, DataStore, Capacity, Length, Buffer);
	HexBytes:= BytesToHex(Buffer, Length, ' ');
	Logger.PrintLine('%s (%d bytes)', [HexBytes, Length]);
	orig_NetSendCommandStore(nil, Session, DataStore);
end;

procedure Hook(pTarget:LPVOID; pDetour:LPVOID; out ppOriginal:LPVOID; DebugName: String);
var
	Status: MH_STATUS;
begin
	Status:= MH_CreateHook(
		pTarget,
		pDetour,
		ppOriginal
	);
	if Status <> MH_OK
		then raise Exception.CreateFmt(
			'Failed to create hook for "%s" function. Code: %d.',
			[DebugName, Status]
		);

	Status:= MH_EnableHook(pTarget);
	if Status <> MH_OK
		then raise Exception.CreateFmt(
			'Failed to enable hook for "%s" function. Code: %d.',
			[DebugName, Status]
		);
end;

procedure InitHooks;
var
	Status: MH_STATUS;
begin
	Status:= MH_Initialize;
	if Status <> MH_OK
		then raise Exception.CreateFmt(
			'Failed to initialize MinHook. Code: %d.',
			[Status]
		);

	Hook(
		_CNetData_FlushCommands,
		@hook_CNetData_FlushCommands,
		Pointer(orig_CNetData_FlushCommands),
		'FlushCommands'
	);
	Hook(
		_NetSendCommandStore,
		@hook_NetSendCommandStore,
		Pointer(orig_NetSendCommandStore),
		'NetSendCommandStore'
	);
end;

procedure InitOffsets;
begin
	try
		GameBase:= GetModule(GetCurrentProcessId, 'Game.dll').modBaseAddr;
	except
		on e: Exception do begin
			e.Message:= 'Failed to find game dll.'#13#10 + e.Message;
			raise;
		end;
	end;
	Pointer(_CNetData_FlushCommands):= GameBase + $54D930;
	Pointer(_NetSendCommandStore):= GameBase + $54D970;
end;

procedure InitLog;
var
	Config: TIniFile;
	LogFilePath: String;
begin
	Config:= TIniFile.Create(CONFIG_PATH);
	try
		LogFilePath:= Config.ReadString('Log', 'Path', '-');

		if LogFilePath = '-'
			then Logger:= TConsoleLogger.Create()
			else Logger:= TFileLogger.Create(LogFilePath);

	finally
		Config.Destroy;
	end;
end;

procedure Init;
begin
	InitLog;
	InitOffsets;
	InitHooks;
end;

procedure Crash(Error: Exception);
begin
	MessageBoxW(
		0,
		PWideChar(UnicodeString(Error.Message)),
		PWideChar(UnicodeString('Error!')),
		MB_ICONERROR
	);
	raise Error;
end;

begin
	try
		Init;
	except
		on e: Exception do Crash(e);
	end;
end.

