unit mylogger;

{$mode ObjFPC}{$H+}

interface

uses
	Classes, SysUtils;

type

	{ TLogger }

	TLogger = class
	public
		procedure Print(AText: String); virtual; abstract;
		procedure Print(ATemplate: String; AParameters: array of const);
		procedure PrintLine(AText: String);
		procedure PrintLine(ATemplate: String; AParameters: array of const);
	end;

	{ TFileLogger }

	TFileLogger = class(TLogger)
	private
		FFileHandle: THandle;
	public
		procedure Print(AText: String); override;
		constructor Create(AFilePath: String);
	end;

	{ TConsoleLogger }

	TConsoleLogger = class(TLogger)
	public
		procedure Print(AText: String); override;
	end;

implementation

const
	LINE_ENDING = #13#10;

{ TLogger }

procedure TLogger.Print(ATemplate: String; AParameters: array of const);
begin
	Print(Format(ATemplate, AParameters));
end;

procedure TLogger.PrintLine(AText: String);
begin
	Print(AText + LINE_ENDING);
end;

procedure TLogger.PrintLine(ATemplate: String; AParameters: array of const);
begin
	Print(ATemplate + LINE_ENDING, AParameters);
end;

{ TConsoleLogger }

procedure TConsoleLogger.Print(AText: String);
begin
	WriteLn(AText);
end;

{ TFileLogger }

procedure TFileLogger.Print(AText: String);
begin
	FileWrite(FFileHandle, AText[1], Length(AText));
end;

constructor TFileLogger.Create(AFilePath: String);
begin
	FFileHandle:= FileCreate(AFilePath, fmShareDenyNone);
	if FFileHandle = THandle(-1)
		then raise Exception.CreateFmt(
			'Failed to create log file "%s".',
			[AFilePath]
		);
end;

end.

